#include "StdAfx.h"
#include "mm.h"
#include "diff_wheel_robot.h"
#include "ParticleFilter.h"

#define MAX_PARTICLE	500

CParticleFilter::CParticleFilter()
{
	_particles.reserve(MAX_PARTICLE*2);

	X.resize (3,1);
	X.null ();

	P.resize (3,3);
	P.null ();
}

CParticleFilter::~CParticleFilter()
{	
	_particles.clear ();
}

void CParticleFilter::predict(double sl, double sr)
{
	// κ ¿  ̵ ŭ  ƼŬ δ.
	//  , ̵ ϴ (Gaussian random  ) ߰ȴ.

	const double kr = 0.035;
	const double kl = 0.035;

	for (unsigned int i = 0; i<_particles.size(); i++) {
		sParticle &s = _particles[i];

		double sln = sl + kl*fabs(sl)*GaussRand ();
		double srn = sr + kr*fabs(sr)*GaussRand ();

		double ds     = (srn + sln)/2;
		double dtheta = (srn - sln)/_axle_length;

		s.x = s.x + ds*cos(s.theta + dtheta/2);
		s.y = s.y + ds*sin(s.theta + dtheta/2);
		s.theta = s.theta + dtheta;
	}
}

double CParticleFilter::probability (sFloorLandmark &l, sParticle &p)
{
	double d = distance (l.x, l.y, p.x, p.y);

	if (d < l.r+0.05) return 1;
	else              return 0.1;
/*
	if (d < r) return 1;
	else if (d < 1.5*r) return 3. - 2.*d/r;
	else return 0.;
*/
}

void CParticleFilter::update(unsigned long long id, double r)
{
	// RFID  RFID ±׸ Ͽ.

	sFloorLandmark *rfid = get_floor_landmark (id);

	assert (rfid);

	if (0 < _particles.size ()) {
		// ƼŬ ±׿ Ÿ  ƼŬ Ȯ Ʈ Ѵ. 
		double sum_w = 0.;

		for(unsigned int i = 0; i<_particles.size(); i++) {
			double p = probability (*rfid, _particles[i]);

			_particles[i].w *= p;

			sum_w += _particles[i].w;
		}
		
		if (sum_w > 0.) {
			//  ƼŬ Ȯ  1 ǵ Ѵ.
			for(unsigned int i = 0; i<_particles.size(); i++) {
				_particles[i].w /= sum_w;
			}

			// ƼŬ ġ  ڽ ƼŬ  ϰ ο ƼŬ Ѵ.
			// , ġ  ƼŬ  ο ƼŬ , ġ  ƼŬ Ҹȴ.
			vector<sParticle> new_particles;
			new_particles.reserve ((int)(MAX_PARTICLE*1.2));

			for(unsigned int i = 0; i<_particles.size(); i++) {
				unsigned int no_particles = INTEGER(_particles[i].w*MAX_PARTICLE);

				for (unsigned int j=0; j<no_particles; ++j) {
					new_particles.push_back(_particles[i]);
				}
			}
			_particles.swap (new_particles);
		}
	}
	else {
		// ƼŬ   . 
		// x, yó ±װ ߰ߵǾ , x, y ߽ ݰ r  ƼŬ Ѹ.
		randomize_at (rfid->x, rfid->y, rfid->r);
	}
}

void CParticleFilter::randomize_at (double x, double y, double radius)
{
	_particles.clear ();
	_particles.reserve (MAX_PARTICLE);

	for (int i=0; i<MAX_PARTICLE; ++i) {
		double r = radius*(rand()%1000)/1000.;
		double theta = _DEG2RAD*(rand()%360);

		_particles.push_back (sParticle (x + r*cos(theta), y + r*sin(theta), theta, 1./MAX_PARTICLE));
	}
}

void CParticleFilter::statistical_pose()
{
	if (_particles.size () <= 0) return;

	X.resize (3,1); X.null();
	P.resize (3,3); P.null();
	
	{
		// (S.x.theta)  Ͽ    ߸  س  ִ.
		//  ڸ, -180 180 Ͽ   0,   180 Ǿ Ѵ.
		//   ذϱ  + ִ  - ִ   ϴµ,
		// -   +  -   Ͽ  Ͽ.
		double wp = 0., wn = 0.;
		double xp[3] = {0, 0, 0};
		double xm[3] = {0, 0, 0};

		for (unsigned int i=0; i<_particles.size (); ++i) {
			sParticle &S = _particles[i];

			double theta = DeltaRad(S.theta, 0.);
			if (0. <= theta) {
				wp += S.w;
				xp[0] += S.w*S.x;
				xp[1] += S.w*S.y;
				xp[2] += S.w*S.theta;
			}
			else {
				wn += S.w;
				xm[0] += S.w*S.x;
				xm[1] += S.w*S.y;
				xm[2] += S.w*S.theta;
			}
		}

		X(0,0)  = (xp[0] + xm[0])/(wp + wn);
		X(1,0)  = (xp[1] + xm[1])/(wp + wn);

		if (0. < wp && 0. < wn) {
			double dth = DeltaRad (xp[2]/wp, xm[2]/wn);
			X(2,0) = xm[2]/wn + dth*wp/(wp + wn);
		}
		else if (0 < wp) {
			X(2,0) = xp[2]/wp;
		}
		else { //if (0 < wn) {
			X(2,0) = xm[2]/wn;
		}
	}

	{
		for (unsigned int i=0; i<_particles.size (); ++i) {
			sParticle &S = _particles[i];

			dMatrix dx (3,1);
			dx(0,0) = S.x - X(0,0);
			dx(1,0) = S.y - X(1,0);
			dx(2,0) = DeltaRad (S.theta, X(2,0));

			P += dx*~dx;
		}
		P /= _particles.size ();
	}
}

