#pragma once
#include "MatrixAlgebra.h"
#include "floor_landmark.h"
#include <vector>

using namespace std;

inline double distance (double x1, double y1, double x2, double y2)
{
	double dx = x2 - x1;
	double dy = y2 - y1;

	return sqrt(dx*dx + dy*dy);
}

struct sParticle
{
	double w;
	double x, y, theta;

	sParticle(double x_, double y_, double theta_, double weight_ = 1.) 
		: x(x_), y(y_), theta(theta_), w(weight_) { }

	sParticle(const sParticle &p)
		: x(p.x), y(p.y), theta(p.theta), w(1.) { }
};

class CParticleFilter
{
public:
	CParticleFilter();
	~CParticleFilter();

	void predict(double sl, double sr);
	void update(unsigned long long id, double r);
	void statistical_pose();
	
	vector<sParticle> _particles;

	dMatrix X, P;

private:
	double probability (sFloorLandmark &l, sParticle &p);
	void randomize_at (double x, double y, double radius);
};